;-----------------------------------------------------------------------------;
; Audio Waveform Monitor R0.2                               May 29, 2005      ;
; (C)ChaN, 2005; http://elm-chan.org/                                         ;
;-----------------------------------------------------------------------------;
;
.include "m8def.inc"
.include "avr.inc"
.include "akiglcd.inc"

.equ	MOSI	= 3	;Port B bit definitions
.equ	LCDCLK	= 1	;/

.equ	RW	= 5	;Port C bit definitions
.equ	E	= 4	;
.equ	CS2	= 3	;
.equ	CS1	= 2	;
.equ	A0	= 1	;
.equ	RES	= 0	;/

.equ	FFT_N	= 128	;Number of samples
.equ	LCD_W	= 122	;LCD size
.equ	LCD_H	= 32	;/
.equ	WAV_W	= (LCD_W-FFT_N/2)

.def	_0	= r15	;Zero
.def	_Flags	= r25	;b0:In captureing
			;b1:In pause
			;b2:MOSI edge detector

;----------------------------------------------------------;
; Data memory area

.dseg
	.org	RAMTOP
CaptBuf:.byte	FFT_N*2		;Sampling buffer
BflyBuf:.byte	FFT_N*4		;Butterfly operation table, Wave form buffer
LvlBuf:	.byte	FFT_N/2		;Spectrum bar length
AdPtr:	.byte	1		;Sampling pointer
LcdDiv:	.byte	1		;Divider for LCD drive


;----------------------------------------------------------;
; Program code area

.cseg
	rjmp	reset		; RESET
	rjmp	0		; INT0
	rjmp	0		; INT1
	rjmp	0		; TC2 COMP
	rjmp	0		; TC2 OVF
	rjmp	0		; TC1 CAPT
	rjmp	0		; TC1 COMPA
	rjmp	0		; TC1 COMPB
	rjmp	0		; TC1 OVF
	rjmp	0		; TC0 OVF
	rjmp	0		; SPI
	rjmp	0		; USART RXC
	rjmp	0		; USART UDRE
	rjmp	0		; USART TXC
;	rjmp	isr_adc		; ADC
;	rjmp	0		; EE_RDY
;	rjmp	0		; ANA_COMP
;	rjmp	0		; TWI
;	rjmp	0		; SPM_RDY


;----------------------------------------------------------;
; ADC interrupt (9.6ksps)


isr_adc:
	push	AL
	in	AL, SREG
	pushw	A
	pushw	Y

	sbrs	_Flags, 0	;Skip if in Idle
	rjmp	adc_skip	;/
	lds	YL, AdPtr	;Write pointer
	clr	YH		;
	addiw	Y, CaptBuf	;/
	inw	A, ADC		;Store A/D value
	subiw	A, 32768	;
	stw	Y+, A		;/
	lds	AL, AdPtr	;Next pointer
	addi	AL, 2		;
	sts	AdPtr, AL	;/
	brne	PC+2		;Terminate if 128 samples captured.
	cbr	_Flags, bit0	;/
adc_skip:

	lds	AL, LcdDiv	;Drive LCD clock (2.4kHz)
	inc	AL		;
	sts	LcdDiv, AL	;
	bst	AL, 1		;
	in	AL, PORTB	;
	bld	AL, LCDCLK	;
	out	PORTB, AL	;/

	popw	Y
	popw	A
	out	SREG, AL
	pop	AL
	reti



;----------------------------------------------------------;
; Initialize peripherals (ATmega8-16AC @16MHz)

reset:
	clr	_0			;Zero reg.
	ldiw	Z, RAMTOP		;Clear all variables
	ldiw	X, RAMEND-RAMTOP+1	;
	st	Z+, _0			;
	sbiw	XL, 1			;
	brne	PC-2			;/
	sbiw	ZL, 1			;SP 
	outw	SP, Z			;/

	outi	PORTB, 0b00101000	;Port B 
	outi	DDRB,  0b00010110

	outi	PORTC, 0b00001100	;Port C 
	outi	DDRC,  0b00111111

	; Start TC1 with 444kHz on OC1B pin (for MAX293 clock)
	ldiw	A, 36-1			;TOP value
	outw	ICR1, A			;/
	ldiw	A, 36/2			;half value
	outw	OCR1B, A		;/
	outi	TCCR1A, 0b00100010	;COM1B[1:0]=10, WGM1[1:0]=10
	outi	TCCR1B, 0b00011001	;WGM1[3:2]=11, CS1[2:0]=001

	; Start A/D with ch6, free running, 9.6ksps
	outi	ADMUX, 0b00100110	;ADLAR=1, MUX[3:0]=0110
	outi	ADCSR, 0b11101111	;ADEN=1,ADSC=1,ADFR=1,ADIE=1,ADPS[2:0]=111

	 rcall	lcd_init

	clr	_Flags
	sei
	sbr	_Flags, bit0	;Start wave form captureing


;----------------------------------------------------------;
; Main


main:
	sbrc	_Flags, 0	;Wait for end of wave captureing
	rjmp	PC-1		;/

	 rcall	make_wave	;Make wave form				[410us]
	 rcall	rfsh_wave	;Display wave form into LCD		[480us]

	 rcall	do_window	;Fill butterfly table			[340us]
	sbr	_Flags, bit0	;Restart captureing for next
	 rcall	do_fft		;Butterfly operations			[4.4ms]
	 rcall	make_bars	;Get scalar values, update spectrum	[3.3ms]
	 rcall	rfsh_bars	;Display spectrum bars into LCD		[660us]

	 rcall	check_pause	;Check MOSI pin for Pause/Resume	[0.8us]

	rjmp	main		; (approx. 70 cycles/sec)



;----------------------------------------------------------;
; Oscilloscope

make_wave:
	ldiw	Z, CaptBuf	;Search zero crossing point
	ldiw	C, FFT_N-WAV_W	;to still cyclic wave form.
	lddw	A, Z+0		;
	ldiw	B, -2000	;
	cpw	A, B		;
	brge	PC+2		;
	sbr	CH, bit0	;
	ldiw	B, 1000		;
	cpw	A, B		;
	brlt	PC+3		;
	sbrc	CH, 0		;
	rjmp	PC+4		;
	adiw	ZL, 2		;
	dec	CL		;
	brne	PC-17		;/

	ldiw	Y, BflyBuf	;Y = Destination
	ldi	CL, WAV_W	;Loop counter
	ldw	A, Z+		;Get a value into A
	addiw	A, 32768	;
	mov	AL, AH		;AL /= 2048 (LCD Y position)
	lsr	AL		;
	lsr	AL		;
	lsr	AL		;/
	st	Y+, AL		;Store value into destinatation
	dec	CL		;Repeat WAV_W times
	brne	PC-10		;/

	ldiw	X, WAV_W*4	;Clear drawing buffer
	st	Y+, _0		;
	sbiw	XL, 1		;
	brne	PC-2		;/
	sbiw	YL, WAV_W	;Drawing start poitnt
	movw	T2L, YL		;/

	ldiw	Z, BflyBuf	;Express waveform into drawing buffer
	ld	AL, Z
	mov	AH, AL
	andi	AL, 7
	clr	BL
	sec
	ror	BL
	subi	AL, 1
	brcc	PC-2
	lsr	AH
	lsr	AH
	lsr	AH
	andi	AH, 3
	ldi	AL, WAV_W
	mul	AL, AH
	movw	YL, T2L
	subw	Y, T0
	ldi	DH, WAV_W-1
dn_l1:	ld	CL, Z+
	ld	CH, Z
	sub	CL, CH
	mov	DL, CL
	brcc	PC+2
	neg	CL
	mov	CH, CL
	lsr	CH
dn_l2:	ld	AL, Y
	or	AL, BL
	st	Y, AL
	subi	CH, 1
	brcc	PC+2
	adiw	YL, 1
	cpi	DL, 0
	breq	PC+11
	brmi	PC+6
	lsl	BL
	brne	PC+8
	ldi	BL, 0x01
	adiw	YL, WAV_W
	rjmp	PC+5
	lsr	BL
	brne	PC+3
	ldi	BL, 0x80
	sbiw	YL, WAV_W
	subi	CL, 1
	brcs	PC+2
	brne	dn_l2
	dec	DH
	brne	dn_l1

	ret


rfsh_wave:
	sbrc	_Flags, 1	;Skip if in pause
	ret			;/
	ldiw	Y, BflyBuf+WAV_W
	clr	BH
	ldi	BL, FFT_N/2
	 rcall	lcd_setadr
	ld	AL, Y+
	 rcall	lcd_put
	cpi	BL, 0
	brne	PC-3
	cpi	BH, 4
	brcs	PC-7
	ret



;----------------------------------------------------------;
; Spectrum analyzer
;----------------------------------------------------------;
; 16bit fixed point FFT performance with megaAVR @16MHz
;
;  Points:  Input,    FFT, Output,   Total: Throughput
;   64pts:  .17ms,  1.9ms,  1.4ms,   3.5ms:   18.3kpps   (expected)
;  128pts:  .34ms,  4.4ms,  2.6ms,   7.3ms:   17.5kpps   (measured)
;  256pts:  .68ms, 10.1ms,  5.2ms,  16.0ms:   16.0kpps   (expected)
;  512pts:  1.4ms, 22.6ms, 10.4ms,  34.4ms:   14.8kpps   (expected)
;
; Input: Input waveform into butterfly table with applying window
; FFT: Execute butterfly operations
; Output: Descramble and output the spectrum as scalar values


do_window: ; Fill butterfly table with applying window function
	ldiw	X, CaptBuf	;Source
	ldiw	Y, BflyBuf	;Destination
	ldiw	Z, t_hamming*2	;Window table
	ldi	CL, FFT_N
ms_l1:	lpmw	A, Z+		;Get a window attenuation ratio
	ldw	B, X+		;Get an input value
	FMULS16	A, B, T4, T2	;Apply window
	stw	Y+, T4		;Store real and image
	stw	Y+, T4		;/
	dec	CL
	brne	ms_l1
	ret


do_fft:	; Execute butterfly operations (not optimized)
	ldi	XH, 1		;u
	ldi	EL, FFT_N/2	;l
df_l1:	ldiw	Z, BflyBuf	;Z = BflyBuf
	ldiw	Y, BflyBuf	;Y = BflyBuf + l * 4
	muli	EL, 4		;
	addw	Y, T0		;/
	mul	AL, XH		;T10L = u * 4
	mov	T10L, T0L	;/
	mov	XL, XH		;w = u
df_l2:	clr	T14L		;p=0
df_l3:	lddw	A, Z+0		;A = [Z+0] - [Y+0], [Z+0] += [Y+0]
	asrw	A		;
	movw	CL, AL		;B = [Z+2] - [Y+2], [Z+2] += [Y+2]
	lddw	D, Y+0		;
	asrw	D		;
	subw	A, D		;
	addw	C, D		;
	stw	Z+, C		;
	lddw	B, Z+0		;
	asrw	B		;
	movw	CL, BL		;
	lddw	D, Y+2		;
	asrw	D		;
	subw	B, D		;
	addw	C, D		;
	stw	Z+, C		;/
	movw	T0L, ZL
	movw	ZL, T14L	;C=cos(p), D=sin(p)
	addiw	Z, t_cos_sin*2	;
	lpmw	C, Z+		;
	lpmw	D, Z+		;/
	movw	ZL, T0L
	FMULS16	A, C, T4, T2	;[Y+0] = A * C + B * D
	FMULS16	B, D, T8, T6	;[Y+2] = B * C - A * D
	addw	T2, T6		;
	adcw	T4, T8		;
	stw	Y+, T4		;
	FMULS16	B, C, T4, T2	;
	FMULS16	A, D, T8, T6	;
	subw	T2, T6		;
	sbcw	T4, T8		;
	stw	Y+, T4		;/
	add	T14L, T10L	;p += u
	rjne	df_l3
	muli	EL, 4		;Y += l * 4, Z += l * 4; (skip split segment)
	addw	Y, T0		;
	addw	Z, T0		;/
	dec	XL		;--w
	rjne	df_l2
	lsl	XH		;u *= 2
	lsr	EL		;l /= 2
	rjne	df_l1
	ret


make_bars: ;Get scalar spectrum and update spectrum bar length
	ldiw	Z, t_desc*2	;Descramble table
	ldiw	Y, LvlBuf	;Bar length buffer
	ldi	CH, FFT_N/2	;Number of elements
ub_l1:	lpmw	A, Z+		;Get a vector value (A = r, B = i)
	ldiw	X, BflyBuf	;
	addw	X, A		;
	ldw	A, X+		;
	ldw	B, X+		;/
	FMULS16	A, A, T4, T2	;A = sqrt(A * A + B * B); 0..32767 (scalar)
	FMULS16	B, B, T8, T6	;
	addw	T2, T6		;
	adcw	T4, T8		;
	SQRT32			;/
	pushw	Z		;CL = log(A)
	ldi	CL, 0		;
	ldiw	Z, logs*2	;
	lpmw	B, Z+		;
	cpw	A, B		;
	brcs	PC+3		;
	inc	CL		;
	rjmp	PC-6		;
	popw	Z		;/
	ld	AH, Y		;Update bar length with peak holding
	subi	AH, 2		; <-peak decay rate
	brcs	PC+3		;
	cp	AH, CL		;
	brcc	PC+2		;
	mov	AH, CL		;
	st	Y+, AH		;/
	dec	CH
	rjne	ub_l1
	ret


rfsh_bars: ;Draw spectrum bars
	sbrc	_Flags, 1	;Skip refresh if in pause.
	ret			;/
	ldi	DL, LCD_H-8+1	; (This routine refreshes LCD with bar length
	ldiw	B, 0		;  list on the LvlBuf[64]. 1st value is DC
	 rcall	lcd_setadr	;  component, 2nd value is fundamental freq
	 rcall	ds_hl		;  and last value is highest freq)
	clr	BL
	inc	BH
	subi	DL, 8
	brcc	PC-5
	ret
ds_hl:	ldiw	Y, LvlBuf
	ldi	CL, FFT_N/2
	ld	AL, Y+
	sub	AL, DL
	brcc	PC+3
	ldi	AH, 0
	rjmp	PC+9
	ldi	AH, 0xFF
	cpi	AL, 8
	brcc	PC+6
	ldi	AH, bit7
	subi	AL, 1
	brcs	PC+3
	asr	AH
	rjmp	PC-3
	mov	AL, AH
	 rcall	lcd_put
	dec	CL
	brne	PC-16
	ret


; These tables must be rebuilt when change FFT_N

t_cos_sin:	; {cos(x),sin(x)} table (0 <= x < pi, in 64 steps)
	.dw	32767, 0, 32727, 1607, 32609, 3211, 32412, 4807
	.dw	32137, 6392, 31785, 7961, 31356, 9511, 30851, 11038
	.dw	30272, 12539, 29621, 14009, 28897, 15446, 28105, 16845
	.dw	27244, 18204, 26318, 19519, 25329, 20787, 24278, 22004
	.dw	23169, 23169, 22004, 24278, 20787, 25329, 19519, 26318
	.dw	18204, 27244, 16845, 28105, 15446, 28897, 14009, 29621
	.dw	12539, 30272, 11038, 30851, 9511, 31356, 7961, 31785
	.dw	6392, 32137, 4807, 32412, 3211, 32609, 1607, 32727
	.dw	0, 32767, -1607, 32727, -3211, 32609, -4807, 32412
	.dw	-6392, 32137, -7961, 31785, -9511, 31356, -11038, 30851
	.dw	-12539, 30272, -14009, 29621, -15446, 28897, -16845, 28105
	.dw	-18204, 27244, -19519, 26318, -20787, 25329, -22004, 24278
	.dw	-23169, 23169, -24278, 22005, -25329, 20787, -26318, 19519
	.dw	-27244, 18204, -28105, 16845, -28897, 15446, -29620, 14009
	.dw	-30272, 12539, -30851, 11038, -31356, 9511, -31784, 7961
	.dw	-32137, 6392, -32412, 4807, -32609, 3211, -32727, 1607

t_hamming: ; Hamming window (for 128 samples)
	.dw	2621, 2639, 2693, 2784, 2910, 3073, 3270, 3502
	.dw	3768, 4068, 4401, 4765, 5161, 5587, 6042, 6525
	.dw	7036, 7571, 8132, 8715, 9320, 9945, 10588, 11249
	.dw	11926, 12616, 13318, 14031, 14753, 15482, 16216, 16954
	.dw	17694, 18433, 19171, 19905, 20634, 21356, 22069, 22772
	.dw	23462, 24138, 24799, 25443, 26068, 26673, 27256, 27816
	.dw	28352, 28862, 29345, 29800, 30226, 30622, 30987, 31319
	.dw	31619, 31885, 32117, 32315, 32477, 32603, 32694, 32748
	.dw	32767, 32748, 32694, 32603, 32477, 32315, 32117, 31885
	.dw	31619, 31319, 30987, 30622, 30226, 29800, 29345, 28862
	.dw	28352, 27816, 27256, 26673, 26068, 25443, 24799, 24138
	.dw	23462, 22772, 22069, 21356, 20634, 19905, 19171, 18433
	.dw	17694, 16954, 16216, 15482, 14753, 14031, 13318, 12616
	.dw	11926, 11249, 10588, 9945, 9320, 8715, 8132, 7571
	.dw	7036, 6526, 6042, 5587, 5161, 4765, 4401, 4068
	.dw	3768, 3502, 3270, 3073, 2910, 2784, 2693, 2639

t_desc:	; Descramble table (for 128 point FFT)
	.dw	0*4, 64*4, 32*4, 96*4, 16*4, 80*4, 48*4, 112*4
	.dw	8*4, 72*4, 40*4, 104*4, 24*4, 88*4, 56*4, 120*4
	.dw	4*4, 68*4, 36*4, 100*4, 20*4, 84*4, 52*4, 116*4
	.dw	12*4, 76*4, 44*4, 108*4, 28*4, 92*4, 60*4, 124*4
	.dw	2*4, 66*4, 34*4, 98*4, 18*4, 82*4, 50*4, 114*4
	.dw	10*4, 74*4, 42*4, 106*4, 26*4, 90*4, 58*4, 122*4
	.dw	6*4, 70*4, 38*4, 102*4, 22*4, 86*4, 54*4, 118*4
	.dw	14*4, 78*4, 46*4, 110*4, 30*4, 94*4, 62*4, 126*4



;----------------------------------------------------------;


check_pause:
	sbic	PINB, MOSI
	rjmp	PC+7
	sbrs	_Flags, 2
	ret
	ldi	AL, bit1
	eor	_Flags, AL
	cbr	_Flags, bit2
	ret
	sbr	_Flags, bit2
	ret



;----------------------------------------------------------;
; LCD control functions (for dual SED1520 based module)


lcd_init:
	ldiw	A, 50000	;12.5ms
	subiw	A, 1		;
	brne	PC-2		;/

	sbi	PORTC, RES	;RES = H
	outi	DDRD, -1	;Set Port D as data port

	ldiw	A, 10000	;2.5ms
	subiw	A, 1		;
	brne	PC-2		;/

	ldiw	B, 0		;Clear all
	 rcall	lcd_setadr	;
	ldi	AL, 0		;
	 rcall	lcd_put		;
	cpi	BH, 4		;
	brcs	PC-3		;/

	cbi	PORTC, CS1	;Enable Display
	cbi	PORTC, CS2	;
	cbi	PORTC, A0	;
	ldi	AL, 0xAF	;
	 rcall	lcd_write	;/

	ret


lcd_setadr:
	push	BL
	clr	AL
	cpi	BL, 61
	brcs	PC+3
	ldi	AL, 2
	subi	BL, 61
	 rcall	lcd_setdst
	mov	AL, BH
	andi	AL, 3
	ori	AL, 0xB8
	 rcall	lcd_write
	mov	AL, BL
	andi	AL, 63
	pop	BL
	rjmp	PC+1
	 rcall	lcd_write
	sbi	PORTC, A0
	rjmp	PC+1
	ret


lcd_put:
	 rcall	lcd_write
	inc	BL
	cpi	BL, 61
	breq	lcd_setadr
	cpi	BL, LCD_W
	brne	PC+4
	inc	BH
	clr	BL
	 rcall	lcd_setadr
	ret


lcd_write:
	out	PORTD, AL
	sbi	PORTC, E
	cbi	PORTC, E
	ret


lcd_setdst:
	in	AH, PORTC
	sbr	AH, (1<<CS1)+(1<<CS2)
	sbrc	AL, 1
	cbr	AH, (1<<CS2)
	sbrs	AL, 1
	cbr	AH, (1<<CS1)
	bst	AL, 0
	bld	AH, A0
	out	PORTC, AH
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	ret

logs:	; 60dB - 32 point log conversion table
	.dw	15, 19, 24, 30, 37, 47, 58, 72, 89, 111, 138, 171, 213, 264, 328, 407, 505, 627, 778, 966, 1199, 1488, 1846, 2292, 2844, 3529, 4380, 5436, 6746, 8371, 10389, 12892, 16000, 0xFFFF

;	for($a = -32; $a <= 0; $a++) {
;		$v = exp($a*1.875*2.303/20);
;		printf("%d, ", $v * 16000)
;	}
;	printf("0xFFFF\n");
